'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Model = exports.clear = exports.get = exports.create = exports.config = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _util = require('./util');

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var SIGNATURE = {};

var getCleanObject = function getCleanObject() {
  return Object.create(null);
};

var createSetter = function createSetter(key) {
  return function set(value) {
    this._data[key] = value;
    delete this[key];
  };
};

var models = getCleanObject();
var globalDefaultRead = true;
var globalDefaultReadFail = null;
var globalDefaultCache = true;

var getModel = function getModel(model) {
  return typeof model === 'string' ? models[model] : model;
};

var createChildModel = function createChildModel(parent, Class, data) {
  return new Class(data, parent._context, parent, parent._root);
};

var createModelMapFn = function createModelMapFn(Class) {
  return function (data) {
    return data && createChildModel(this, Class, data);
  };
};

var createSimpleGetter = function createSimpleGetter(key) {
  return function () {
    return this._data[key];
  };
};

var wrapGetterWithModel = function wrapGetterWithModel(key, getter, _ref) {
  var type = _ref.type;
  var list = _ref.list;

  if (!type) {
    return getter;
  }

  var _mapFn = void 0;
  if (typeof type === 'string') {
    _mapFn = function mapFn(data) {
      if (!models[type]) {
        throw new Error('Unknown field model. model=\'' + type + '\'');
      }
      _mapFn = createModelMapFn(models[type]);
      return _mapFn.call(this, data);
    };
  } else {
    _mapFn = createModelMapFn(type);
  }

  if (list) {
    // List type field.
    return function get() {
      var val = getter.call(this);
      return val && val.map(_mapFn, this);
    };
  }

  // Non-list type field.
  return function get() {
    return _mapFn.call(this, getter.call(this));
  };
};

var wrapGetterWithAccess = function wrapGetterWithAccess(key, getter, _ref2) {
  var read = _ref2.read;
  var readFail = _ref2.readFail;

  if (read === true) {
    return getter;
  }

  if (!read) {
    if (typeof readFail === 'function') {
      return function get() {
        return readFail(this, key);
      };
    }
    return function get() {
      return readFail;
    };
  }

  if (typeof readFail === 'function') {
    return function get() {
      if (read(this, key)) {
        return getter.call(this);
      }
      return readFail(this, key);
    };
  }

  return function get() {
    if (read(this, key)) {
      return getter.call(this);
    }
    return readFail;
  };
};

var wrapGetterWithCache = function wrapGetterWithCache(key, getter, _ref3) {
  var _ref3$cache = _ref3.cache;
  var cache = _ref3$cache === undefined ? globalDefaultCache : _ref3$cache;

  if (!cache) {
    return getter;
  }

  return function get() {
    var value = getter.call(this);
    this[key] = value;
    (0, _util.setProperty)(this, key, value);
    return value;
  };
};

var createGetter = function createGetter(prototype, key, rule) {
  var getter = void 0;
  getter = createSimpleGetter(key);
  getter = wrapGetterWithModel(key, getter, rule);
  getter = wrapGetterWithAccess(key, getter, rule);
  getter = wrapGetterWithCache(key, getter, rule);
  return getter;
};

// exports

var config = exports.config = function config(defaults) {
  if ('read' in defaults) globalDefaultRead = defaults.read;
  if ('readFail' in defaults) globalDefaultReadFail = defaults.readFail;
  if ('cache' in defaults) globalDefaultCache = Boolean(defaults.cache);
};

var create = exports.create = function create() {
  var _ref4 = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var name = _ref4.name;
  var _ref4$base = _ref4.base;
  var base = _ref4$base === undefined ? Model : _ref4$base;
  var _ref4$props = _ref4.props;
  var props = _ref4$props === undefined ? getCleanObject() : _ref4$props;
  var _ref4$rules = _ref4.rules;
  var rules = _ref4$rules === undefined ? getCleanObject() : _ref4$rules;
  var _ref4$interfaces = _ref4.interfaces;
  var interfaces = _ref4$interfaces === undefined ? [] : _ref4$interfaces;

  var NewModel = function (_base) {
    _inherits(NewModel, _base);

    function NewModel() {
      _classCallCheck(this, NewModel);

      return _possibleConstructorReturn(this, Object.getPrototypeOf(NewModel).apply(this, arguments));
    }

    return NewModel;
  }(base);

  // name
  if (models[name]) {
    throw new Error('\'' + name + '\' model already exists!');
  }
  models[name] = NewModel;
  (0, _util.defineClassName)(NewModel, name);

  // signature
  (0, _util.defineStatic)(NewModel, '$signature', SIGNATURE);

  // props

  var Props = function Props(model) {
    _classCallCheck(this, Props);

    this.$model = model;
  };

  (0, _util.defineStatic)(NewModel, '$Props', Props);

  (0, _util.forEach)(props, function (fn, key) {
    return (0, _util.defineLazyProperty)(Props.prototype, key, function () {
      return fn.call(this, this.$model);
    });
  });

  (0, _util.defineLazyProperty)(NewModel.prototype, '$props', function () {
    return new Props(this);
  });

  // interfaces
  (0, _util.defineStatic)(NewModel, '$interfaces', interfaces);

  if (base !== Model) {
    (0, _util.inheritClass)(NewModel, base);
    if (base.$Props) (0, _util.inheritClass)(Props, base.$Props);
  }

  interfaces.forEach(function (from) {
    (0, _util.inheritClass)(NewModel, from);
    if (from.$Props) (0, _util.inheritClass)(Props, from.$Props);
  });

  // rules
  var _rules$$default = rules.$default;
  _rules$$default = _rules$$default === undefined ? {} : _rules$$default;
  var _rules$$default$read = _rules$$default.read;
  var defaultRead = _rules$$default$read === undefined ? globalDefaultRead : _rules$$default$read;
  var _rules$$default$readF = _rules$$default.readFail;
  var defaultReadFail = _rules$$default$readF === undefined ? globalDefaultReadFail : _rules$$default$readF;

  var otherRules = _objectWithoutProperties(rules, ['$default']);

  (0, _util.forEach)(otherRules, function (rule, key) {
    if (rule === true) {
      rule = {
        read: true,
        readFail: defaultReadFail
      };
    } else if (rule === false) {
      rule = {
        read: false,
        readFail: defaultReadFail
      };
    } else if (typeof rule === 'function') {
      rule = {
        read: rule,
        readFail: defaultReadFail
      };
    } else {
      if (rule.read === undefined) rule.read = defaultRead;
      if (rule.readFail === undefined) rule.readFail = defaultReadFail;
    }

    (0, _util.defineGetterSetter)(NewModel.prototype, key, createGetter(NewModel.prototype, key, rule), createSetter(key));
  });

  return NewModel;
};

var get = exports.get = getModel;

var clear = exports.clear = function clear() {
  models = getCleanObject();
};

var Model = exports.Model = function () {
  function Model(data, context, parent, root) {
    _classCallCheck(this, Model);

    this._data = data;
    this._parent = parent || null;
    this._root = root || this;
    this._context = context;
  }

  _createClass(Model, [{
    key: '$destroy',
    value: function $destroy() {
      delete this._data;
      delete this._parent;
      delete this._context;
    }
  }, {
    key: '$get',
    value: function $get(name) {
      return this._data[name];
    }
  }, {
    key: '$parentOfType',
    value: function $parentOfType(type) {
      var Model = getModel(type);
      var p = this;
      while (p = p._parent) {
        if (p instanceof Model) {
          return p;
        }
      }
      return null;
    }
  }, {
    key: '$createChild',
    value: function $createChild(model, data) {
      return createChildModel(this, getModel(model), data);
    }
  }, {
    key: '$createChildren',
    value: function $createChildren(model, list) {
      var _this2 = this;

      var Class = getModel(model);
      return list && list.map(function (data) {
        return createChildModel(_this2, Class, data);
      });
    }
  }, {
    key: '$implements',
    value: function $implements(Type) {
      return this.constructor.$interfaces.indexOf(Type) >= 0;
    }
  }, {
    key: '$data',
    get: function get() {
      return this._data;
    }
  }, {
    key: '$parent',
    get: function get() {
      return this._parent;
    }
  }, {
    key: '$context',
    get: function get() {
      return this._context;
    }
  }, {
    key: '$root',
    get: function get() {
      return this._root;
    }
  }]);

  return Model;
}();

exports.default = {
  config: config,
  create: create,
  get: get,
  clear: clear,
  Model: Model
};