'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.forEach = forEach;
exports.setProperty = setProperty;
exports.defineStatic = defineStatic;
exports.defineMethod = defineMethod;
exports.defineLazyProperty = defineLazyProperty;
exports.defineGetterSetter = defineGetterSetter;
exports.inheritPropertyFrom = inheritPropertyFrom;
exports.inheritFrom = inheritFrom;
exports.inheritStatic = inheritStatic;
exports.inheritPrototype = inheritPrototype;
exports.inheritClass = inheritClass;

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function forEach(obj, fn) {
  Object.keys(obj).forEach(function (key) {
    return fn(obj[key], key, obj);
  });
}

function setProperty(obj, key, value) {
  return Object.defineProperty(obj, key, {
    value: value,
    enumerable: true,
    writable: true,
    configurable: true
  });
}

function defineStatic(Class, name, value) {
  return Object.defineProperty(Class, name, {
    value: value,
    writable: false,
    enumerable: false,
    configurable: true
  });
}

function defineMethod(prototype, name, value) {
  Object.defineProperty(prototype, name, {
    value: value,
    writable: true,
    enumerable: false,
    configurable: true
  });
}

function defineLazyProperty(obj, name, fn) {
  var _ref = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];

  var _ref$writable = _ref.writable;
  var writable = _ref$writable === undefined ? true : _ref$writable;
  var _ref$enumerable = _ref.enumerable;
  var enumerable = _ref$enumerable === undefined ? true : _ref$enumerable;
  var _ref$configurable = _ref.configurable;
  var configurable = _ref$configurable === undefined ? true : _ref$configurable;

  Object.defineProperty(obj, name, {
    get: function get() {
      // Use 'this' instead of obj so that obj can be a prototype.
      var value = fn.call(this);
      Object.defineProperty(this, name, {
        value: value,
        writable: writable,
        enumerable: enumerable,
        configurable: configurable
      });
      return value;
    },

    enumerable: enumerable,
    configurable: true
  });
}

var defineClassName = void 0;
if (function () {
  var A = function A() {
    _classCallCheck(this, A);
  };
  try {
    defineStatic(A, 'name', 'B');
    return A.name === 'B';
  } catch (e) {
    return false;
  }
}()) {
  exports.defineClassName = defineClassName = function defineClassName(Class, value) {
    return defineStatic(Class, 'name', value);
  };
} else {
  // Old Node versions require the following options to overwrite class name.
  exports.defineClassName = defineClassName = function defineClassName(Class, value) {
    return Object.defineProperty(Class, 'name', {
      value: value,
      writable: false,
      enumerable: false,
      configurable: false
    });
  };
}
exports.defineClassName = defineClassName;
function defineGetterSetter(Class, name, get, set) {
  return Object.defineProperty(Class, name, {
    get: get,
    set: set,
    enumerable: true,
    configurable: true
  });
}

function inheritPropertyFrom(objA, objB, key, asKey) {
  return Object.defineProperty(objA, asKey || key, Object.getOwnPropertyDescriptor(objB, key));
}

function inheritFrom(objA, objB, excludes) {
  var aKeys = Object.getOwnPropertyNames(objA);
  var bKeys = Object.getOwnPropertyNames(objB);

  var keys = bKeys.filter(function (key) {
    return aKeys.indexOf(key) === -1;
  });
  if (excludes) {
    keys = keys.filter(function (key) {
      return excludes.indexOf(key) === -1;
    });
  }

  keys.forEach(function (key) {
    return inheritPropertyFrom(objA, objB, key);
  });
  return objA;
}

function inheritStatic(classA, classB) {
  inheritFrom(classA, classB, ['length', 'name', 'prototype']);
  return classA;
}

function inheritPrototype(classA, classB) {
  inheritFrom(classA.prototype, classB.prototype, ['constructor']);
  return classA;
}

function inheritClass(classA, classB) {
  inheritStatic(classA, classB);
  inheritPrototype(classA, classB);
  return classA;
}